/*-----------------------------------------------------------------------------------
$Id: test_regression.c,v 1.4 2012/03/06 08:54:23 wolf Exp wolf $
-----------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "regression.h"

#define	PI	(3.1415926535)
#define	RHO	(180.0/3.1415926535)
#define EARTH	(40.e6/360.0)

/*=================================================================
**	Verwendet printf(%f)
**	Nur zum Test auf der Entwicklungsmaschine, nicht auf dem ARM!
**===============================================================*/

/*=================================================================
**	Structs zur Abstraktion
**===============================================================*/

typedef struct {
	double	lat;
	double	lon;
} t_geo;

typedef struct {
	double	north;
	double	east;
} t_rectangular;

typedef struct {
	double	delta;
	double	alpha;	// radian
} t_polar;

/*=================================================================
**	Umrechnungen
**===============================================================*/

void
geodiff (t_geo * von, t_geo* nach, t_geo* delta) {
	delta->lat = nach->lat - von->lat;
	delta->lon = nach->lon - von->lon;
}

void
geo2rect (t_geo * g, t_rectangular* r, double lat) {
	r->north= g->lat * EARTH;
	r->east = g->lon * EARTH * cos (lat/RHO);
}

void
rect2polar (t_rectangular* r, t_polar* p) {
	p->delta = sqrt (r->north*r->north + r->east*r->east);
	p->alpha = atan2 (r->east, r->north);
}

void
polar2rect (t_rectangular* r, t_polar* p) {
	r->east = sin (p->alpha) * p->delta;
	r->north= cos (p->alpha) * p->delta;
}

int
radian2deg (double alpha) {
	return (int) (alpha * RHO + 3600.5) % 360;
}

/*=================================================================
**	Regression für Koordinaten (lat+lon)
**===============================================================*/

typedef struct {
	t_regression	lat;
	t_regression	lon;
} t_georeg;

void
georegInit (t_georeg* reg, int size, t_regression_xy * buffer1, t_regression_xy * buffer2) {
	regressionInit (&reg->lat, size, buffer1);
	regressionInit (&reg->lon, size, buffer2);
}

void
georegAdd (t_georeg* reg, double time, t_geo* geo) {
	regressionAdd (&reg->lat, time, geo->lat);
	regressionAdd (&reg->lon, time, geo->lon);
}

void
georegActual (t_georeg* reg, t_geo* geo) {
	geo->lat = regressionActual (&reg->lat);
	geo->lon = regressionActual (&reg->lon);
}

void
georegChange (t_georeg* reg, t_geo* geo) {
	geo->lat = regressionChange (&reg->lat);
	geo->lon = regressionChange (&reg->lon);
}

/*=================================================================
**	FAKE GPS Empfaenger
**===============================================================*/

float
rnd() { /* -1 .. +1 */
	return (rand() - 0.5 * RAND_MAX) / (RAND_MAX/2);
}

void
readGps (t_geo * gps) {
	static int i;
	float Rauschstaerke = 3; 
	++i;
	gps->lat = 47.0 + 1.0e-5 * (i+Rauschstaerke*rnd());
	gps->lon = 14.0 + 1.4e-5 * (i+Rauschstaerke*rnd());
}

/*=================================================================
**
**	Jede Regression braucht einen eigenen Wertespeicher
**
**	Beim Initialisieren der Regression wird ein Zeiger auf den
**	Wertespeicher und die Anzahl der Stützstellen übergeben.
**
**	Die Anzahl der Stützstellen DARF NICHT größer sein
**	als der Wertespeicher
**
**===============================================================*/

static	t_regression_xy buffer1[50];
static	t_regression_xy buffer2[50];
static	t_regression_xy buffer3[50];

//----------------------------------------------
// 1. Messung (wer misst, misst Mist)
//----------------------------------------------

double		jetzt;			// Sekundenuhr
t_geo		gps;			// gemessene verrauschte Position

//----------------------------------------------
// 2. Glättung von Position und Geschwindigkeit
//----------------------------------------------

t_georeg	georeg;			// Regression fuer Istposition
t_geo		ist;			// Ist-Position aus Regression
t_geo		speed;			// Ist Geschwindigkeit in Grad/s

//----------------------------------------------
// 3. Geschwindigkeit->Polar und Grad
//----------------------------------------------

t_rectangular	mspeed;			// Ist Geschwindigkeit in m/s (N/E)
t_polar		pspeed;			// Ist Geschwindigkeit in m/s + Winkel
int		marsch;			// Marschrichtung

//----------------------------------------------
// 4. Sollposition + Entfernung zum Ziel
//----------------------------------------------

t_geo		soll;			// Soll-Position
t_geo		delta;			// Entfernung zum Ziel (in Grad)

//----------------------------------------------
// 5. Entfernung->Polar und Grad
//----------------------------------------------

t_rectangular	rdelta;			// Entfernung zum Ziel (metrisch N/E)
t_polar		pdelta;			// Entfernung zum Ziel (metrisch+Winkel)
int		peil;			// Peilung zum Ziel

//----------------------------------------------
// 6. Winkel zwischen Marschr. und Peilung
//----------------------------------------------

int		winkel;			// Winkel auf Anzeige von Marsch zum Ziel

//----------------------------------------------
// 7. Entfernung zum Ziel -> Geschw. zum Ziel
//----------------------------------------------

t_regression 	deltareg;		// Glättung und Geschwindigkeit
double		estimatedTime;		// Zeit bis zum Ziel

int main () {

	int i;

	/*=================================================================
	**
	**	Beim Initialisieren der Regression wird ein Zeiger auf den
	**	Wertespeicher und die Anzahl der Stützstellen übergeben.
	**
	**	Die Anzahl der Stützstellen DARF NICHT größer sein
	**	als der Wertespeicher
	**
	**	Die Initialisierung das MEHRFACH aufgerufen werden,
	**	auch mit abweichender ANzahl der Stützstellen.
	**
	**===============================================================*/

	georegInit (&georeg, 40, buffer1, buffer2);
	regressionInit (&deltareg, 20, buffer3);

	/**************************************************************
	**	Sollwerte für Entfernung zum Ziel
	**************************************************************/

	soll.lat = 47.0007;
	soll.lon = 14.0010;

	for (i=0; i<=100; i++) {

		if (i%10==0)
		printf ("Zeit lat       lon         latSpeed  lonSpeed   Geschwindigkeit      Entfernung  Zeiger Restzeit\n");

		/**************************************************************
		**	Messen
		**************************************************************/

		jetzt = (double) i;
		readGps		(&gps);				// lese GPS (mit Rauschen)

		/**************************************************************
		**	Rechnen
		**************************************************************/

		georegAdd	(&georeg, jetzt, &gps);		// füttere die Regression
		georegActual	(&georeg, &ist);		// lese gemittelte Position
		georegChange	(&georeg, &speed);		// lese gemittelte Geschwindigkeit
		geo2rect	(&speed, &mspeed, ist.lat);	// konvertiere nach m/s (je N/E)
		rect2polar	(&mspeed, &pspeed);		// konvertiere in Richtung und Geschw.
		marsch = radian2deg (pspeed.alpha);

		/**************************************************************
		**	Entfernung zum Ziel
		**************************************************************/

		geodiff		(&ist, &soll, &delta);		// Entfernung in Grad (N/E)
		geo2rect	(&delta, &rdelta, ist.lat);	// Entfernung in m (je N/E)
		rect2polar	(&rdelta, &pdelta);		// Entfernung in m und Grad
		peil	= radian2deg (pdelta.alpha);

		winkel	= radian2deg (pdelta.alpha-pspeed.alpha);	// Winkel gegenüber Marschrichtung

		/**************************************************************
		**	Geschwindigkeit und Zeit bis zum Ziel
		**************************************************************/

		regressionAdd (&deltareg, jetzt, pdelta.delta);	// füttere die Regression
		double	eta = regressionReverse (&deltareg, 0);	// Ankunftszeit (bei akt. geschwindigkeit)
								// bei Entfernen vom Ziel auch in der Vergangenheit
		double	rest= eta - jetzt;			// Restzeit = Ankunftszeit - jetzt

		//-----------------------------------------------------------------------------
		// Wenn Geschwindigkeit zu gering oder Zeit in der Vergangenheit, keine Anzeige
		//-----------------------------------------------------------------------------

		if (abs(regressionChange (&deltareg))<0.5 || rest < 5.0) rest = 0.0;

		/**************************************************************
		**	Anzeigen
		**************************************************************/

		printf ("%4i %9.6f %9.6f  %9.6f %9.6f  %9.6fm/s %03i  %9.1fm %03i   <%03i>   %6.3f\n",
			i,
			ist.lat, ist.lon,
			speed.lat, speed.lon,
			pspeed.delta, marsch,
			pdelta.delta, peil,
			winkel,
			rest);
	}
	return 0;
}

/*-----------------------------------------------------------------------------------
$Id: test_regression.c,v 1.4 2012/03/06 08:54:23 wolf Exp wolf $
-----------------------------------------------------------------------------------*/
